@@ -9,26 +9,34 @@ from typing import Awaitable
 from typing import Callable
 from typing import Optional
 
-from .exceptions import *
 from .extension import KCP
 
 
-@dataclass(slots=True)
+DataMutator = Callable[[bytes], bytes]
+
+
+@dataclass()
 class Connection:
     _kcp: KCP
     _server: KCPServerAsync
     address: str
     port: int
-    last_active: float = 0
+    last_active: float
+    _data_mutators: list[DataMutator]
 
     def __post_init__(self) -> None:
         self._kcp.include_outbound_handler(self._send_kcp)
-        self.last_active = time.perf_counter()
 
     @property
     def address_tuple(self) -> AddressType:
         return self.address, self.port
 
+    def _perform_mutations(self, data: bytes) -> bytes:
+        for mutator in self._data_mutators:
+            data = mutator(data)
+
+        return data
+
     def receive(self, data: bytes) -> None:
         """Handles receiving data from the client."""
         self._kcp.receive(data)
@@ -36,9 +44,8 @@ class Connection:
         self.last_active = time.perf_counter()
 
         for data in self._kcp.get_all_received():
-            self._server._loop.create_task(
-                self._server._data_handler(self, data),  # type: ignore
-            )
+            data = self._perform_mutations(data)
+            await self._server._data_handler(self, data)  # type: ignore
 
     def enqueue(self, data: bytes) -> None:
         """Enqueues data to be sent to the client."""
@@ -61,8 +68,17 @@ class Connection:
         """
         self._kcp.update(ts_ms)
 
+    # Decorators
+    def add_data_mutator(self, mutator: DataMutator) -> None:
+        """Adds a local data mutator to the connection.
 
-AddressType = tuple[str | Any, int]
+        :param mutator: The mutator to add.
+        """
+
+        self._data_mutators.append(mutator)
+
+
+AddressType = tuple[Any, int]
 DataHandler = Callable[[Connection, bytes], Awaitable[None]]
 EventHandler = Callable[[], Awaitable[None]]
 
@@ -101,14 +117,14 @@ class KCPServerAsync(asyncio.DatagramProtocol):
         self._on_start: Optional[EventHandler] = None
         self._on_stop: Optional[EventHandler] = None
 
+        self._data_mutators: list[DataMutator] = []
+
     # Private API
     # Called by the protocol.
-    def _handle_data(self, data: bytes, address: AddressType) -> None:
+    def datagram_received(self, data: bytes, address: AddressType) -> None:
         connection = self._ensure_connection(address)
         connection.receive(data)
 
-    datagram_received = _handle_data  # type: ignore
-
     def _ensure_connection(self, address: AddressType) -> Connection:
         connection = self._connections.get(address)
         if connection is None:
@@ -117,6 +133,8 @@ class KCPServerAsync(asyncio.DatagramProtocol):
                 _server=self,
                 address=address[0],
                 port=address[1],
+                last_active=time.perf_counter(),
+                _data_mutators=self._data_mutators.copy(),
             )
             self._connections[address] = connection
 
@@ -145,7 +163,6 @@ class KCPServerAsync(asyncio.DatagramProtocol):
             for connection in tuple(self._connections.values()):
                 connection.update(current_time_ms)
 
-                # Cleaning up unused connections.
                 if current_time - connection.last_active > self.connection_timeout:
                     self._connections.pop(connection.address_tuple)
 
@@ -184,3 +201,9 @@ class KCPServerAsync(asyncio.DatagramProtocol):
         stops listening for connections."""
         self._on_stop = handler
         return handler
+
+    def add_data_mutator(self, mutator: DataMutator) -> DataMutator:
+        """Decorator registering a global mutator which will be called on all data
+        received from a client."""
+        self._data_mutators.append(mutator)
+        return mutator
